@@ -26,6 +26,8 @@ os.environ['NUMEXPR_NUM_THREADS'] = '1'
 
 # Import the DNS cache
 from insyt_secure.utils.dns_cache import DNSCache
+from insyt_secure.utils.audit_logger import AuditLogger
+from insyt_secure.config import settings
 
 # Configure module-level logger
 logger = logging.getLogger(__name__)
@@ -65,7 +67,7 @@ class CodeExecutor:
     """
     def __init__(self, mqtt_broker, mqtt_port, mqtt_username, mqtt_password, subscribe_topic, 
                  publish_topic=None, ssl_enabled=False, allowed_ips=None, always_allowed_domains=None, 
-                 max_workers=5, execution_timeout=30):
+                 max_workers=20, execution_timeout=30, project_id=None):
         """
         Initialize the CodeExecutor with connection details.
         
@@ -83,6 +85,7 @@ class CodeExecutor:
             always_allowed_domains: Domains that are always allowed regardless of IP restrictions
             max_workers: Maximum number of concurrent executions (default: 5)
             execution_timeout: Maximum execution time in seconds for each code snippet (default: 30)
+            project_id: Optional project identifier for context in execution environment
         """
         # Connection parameters - must be provided explicitly
         self.mqtt_broker = mqtt_broker
@@ -99,6 +102,7 @@ class CodeExecutor:
         self.pod_name = os.getenv('POD_NAME', 'local-executor')
         self.max_workers = max_workers
         self.execution_timeout = execution_timeout  # Store default timeout
+        self.project_id = project_id  # Store project context for execution environment
         
         # MQTT message tracking for diagnostics
         self.message_received_count = 0
@@ -133,6 +137,22 @@ class CodeExecutor:
         self.dns_cache = DNSCache(ttl_seconds=86400)  # 24 hours
         logger.info("DNS cache initialized with 24-hour TTL for resilience against DNS outages")
         
+        # Initialize audit logger if enabled
+        self.audit_logger = None
+        if settings.AUDIT_LOGGING_ENABLED:
+            try:
+                self.audit_logger = AuditLogger(
+                    db_path=settings.AUDIT_DB_PATH,
+                    max_size_gb=settings.AUDIT_MAX_SIZE_GB,
+                    max_retention_days=settings.AUDIT_MAX_RETENTION_DAYS
+                )
+                logger.info("Audit logging enabled - execution logs will be stored")
+            except Exception as e:
+                logger.error(f"Failed to initialize audit logger: {e}")
+                logger.warning("Continuing without audit logging")
+        else:
+            logger.info("Audit logging disabled")
+        
         # Validate host against allowed IPs
         if self.allowed_ips:
             logger.debug(f"Network restrictions enabled")
@@ -889,7 +909,7 @@ class CodeExecutor:
             # Log request with masked ID and current timestamp
             masked_id = request_id[-4:] if len(request_id) > 4 else "****"
             current_time = time.strftime("%Y-%m-%d %H:%M:%S")
-            logger.info(f"[{current_time}] Processing execution request (ID: ****{masked_id}, length: {code_length} chars)")
+            logger.info(f"[{current_time}] Processing execution request (ID: ****{masked_id}, input length: {code_length} chars)")
             logger.debug(f"Using timeout: {execution_timeout}s for request ID: {request_id}")
             
             # Execute the code with the specified timeout
@@ -924,6 +944,12 @@ class CodeExecutor:
             
             # Log execution time
             actual_execution_time = time.time() - execution_start_time
+            # Log output character count
+            output_length = len(result) if result is not None else 0
+            if output_length == 0:
+                logger.info(f"[****{masked_id}] Output is empty or null (output length: 0 chars)")
+            else:
+                logger.info(f"[****{masked_id}] Output length: {output_length} chars")
             
             # Determine if execution was successful (no error message in result)
             has_error = isinstance(result, str) and ("Error" in result or "timed out" in result or "cancelled" in result)
@@ -954,7 +980,7 @@ class CodeExecutor:
             if error_description:
                 response["errorDescription"] = error_description
             
-            # Publish response
+            # Publish response FIRST (time-sensitive)
             logger.debug(f"Publishing result with status: {status}")
             # Log detailed info only at debug level
             logger.debug(f"Publishing to channel: {response_topic}")
@@ -992,6 +1018,32 @@ class CodeExecutor:
                     logger.error(f"Exception during publish: {str(e)}")
             else:
                 logger.error(f"Cannot publish result: Not connected to signaling server")
+            
+            # Log execution to audit database AFTER publishing (non-blocking for response time)
+            if self.audit_logger:
+                try:
+                    # Extract query, user, and group from payload if present
+                    query = payload.get("query", "No query provided")
+                    user = payload.get("user", "unknown")
+                    group = payload.get("group")
+                    
+                    # Schedule audit logging in background to not block
+                    asyncio.create_task(
+                        asyncio.get_event_loop().run_in_executor(
+                            None,  # Use default executor
+                            self.audit_logger.log_execution,
+                            query,
+                            python_code,
+                            user,
+                            status,
+                            group,
+                            result if result else None,
+                            error_description
+                        )
+                    )
+                except Exception as audit_error:
+                    # Don't let audit logging failures affect execution
+                    logger.error(f"Failed to schedule audit log: {audit_error}")
                 
         except Exception as e:
             # Get total processing time regardless of errors
@@ -1364,12 +1416,41 @@ class CodeExecutor:
         except ImportError as e:
             logger.warning(f"Could not import safe module: {str(e)}")
         
-        return {
-            '__builtins__': safe_builtins,
-            **safe_modules
+        # Add special Python variables that are normally available
+        special_vars = {
+            # Standard Python special variables
+            '__name__': '__main__',  # Standard value for executed scripts
+            '__file__': '<string>',  # Indicate this is executed from string
+            '__doc__': None,         # No docstring for dynamic code
+            '__package__': None,     # Not part of a package
+            '__spec__': None,        # No module spec for dynamic code
+            '__annotations__': {},   # Empty annotations dict
+            '__cached__': None,      # No cached bytecode for dynamic code
+            '__loader__': None,      # No loader for dynamic code
+            '__debug__': __debug__,  # Python's debug flag (inherits from runtime)
+            
+            # Insyt-specific execution context variables
+            '__insyt_executor__': 'secure_code_executor',  # Execution environment identifier
+            '__execution_mode__': 'sandboxed',             # Execution mode indicator
+            '__security_level__': 'restricted',            # Security level indicator
+            '__data_science_mode__': True,                 # Indicates data science capabilities
+            
+            # Version and context information (helpful for debugging)
+            '__executor_version__': '0.4.2',              # From pyproject.toml version
+            '__python_version__': f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
+            
+            # Project context (set dynamically if available)
+            '__project_id__': self.project_id,            # Current project context
+            '__execution_id__': str(uuid.uuid4())[:8],    # Unique execution identifier
         }
 
-    def extract_and_run_python_code(self, code):
+        return {
+            '__builtins__': safe_builtins,
+            **safe_modules,
+            **special_vars
+        }
+
+    def extract_and_run_python_code(self, code, start_time=None):
         """Execute code and capture its output."""
         try:
             logger.debug("Starting code execution")
@@ -1445,8 +1526,11 @@ class CodeExecutor:
 
             # Only show a preview of the output
             result_preview = (result[:50] + "...") if len(result) > 50 else result
-            execution_time = time.time() - getattr(self, 'execution_start_time', time.time())
-            
+            # Use the provided start_time if available, else use current time
+            if start_time is not None:
+                execution_time = time.time() - start_time
+            else:
+                execution_time = 0.0
             # Combined log message for execution completion
             if execution_success:
                 logger.info(f"Code execution completed successfully in {execution_time:.2f}s")
@@ -1478,18 +1562,15 @@ class CodeExecutor:
         """Execute code with a timeout."""
         try:
             logger.debug(f"Running code with {timeout}s timeout")
-            # Store the execution start time for accurate timing in the extract_and_run_python_code method
-            self.execution_start_time = time.time()
-            
+            start_time = time.time()
             # Set up the timeout mechanism
             with concurrent.futures.ThreadPoolExecutor() as executor:
-                future = executor.submit(self.extract_and_run_python_code, code_block)
+                future = executor.submit(self.extract_and_run_python_code, code_block, start_time)
                 try:
                     result, parsed_result = future.result(timeout=timeout)
                     return result, parsed_result
                 except concurrent.futures.TimeoutError:
                     logger.warning(f"Execution timed out after {timeout}s")
-                    # Attempt to cancel the future to clean up resources
                     if not future.done():
                         future.cancel()
                         logger.info("Cancelled timed-out execution task")
@@ -1730,7 +1811,14 @@ class MQTTMessageWrapper:
         self.mid = mqtt_message.mid
 
 async def main():
-    executor = CodeExecutor()
+    # This is a test function - normally CodeExecutor would be created with proper parameters
+    executor = CodeExecutor(
+        mqtt_broker="localhost",
+        mqtt_port=1883,
+        mqtt_username="test",
+        mqtt_password="test",
+        subscribe_topic="test/topic"
+    )
     await executor.start()
 
 if __name__ == "__main__":
